--[[
	Here I do the combatlog stuff. 
	
	Todo: 
		SPELL_AURA_REMOVED_DOSE
		SPELL_DISPEL
		SPELL_STOLEN
]]
local folder, core = ...

local GetTime = GetTime
local _G = _G
local UnitGUID = UnitGUID
local table_getn = table.getn
local table_remove = table.remove
local table_insert = table.insert
local bit_band = bit.band
local COMBATLOG_OBJECT_TYPE_PLAYER = COMBATLOG_OBJECT_TYPE_PLAYER
local nametoGUIDs = core.nametoGUIDs
local type = type



if not core.LibNameplate then return end
local LibNameplate = core.LibNameplate


local P
local spellInfo
local playerGUID
local Debug = core.Debug
local guidBuffs = core.guidBuffs
local _ --underscore so GetGlobals doesn't nag me.

local prev_OnEnable = core.OnEnable
function core:OnEnable()
	prev_OnEnable(self)
	P = self.db.profile
	spellInfo = self.db.global.spellInfo
	
	playerGUID = UnitGUID("player")

	if P.watchCombatlog == true then
		self:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
	end
end

----------------------------------------------------------------------
function core:COMBAT_LOG_EVENT_UNFILTERED(event, ...)				--
-- Combatlog event handler. 										--
-- Check if we have eventType function then pass the event to it.	--
----------------------------------------------------------------------
	--timestamp, eventType, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags
	local _, eventType  = ...-- ***
--~ 	Debug(event, eventType, self[eventType])

--~ 	Debug(event, "eventType: "..tostring(eventType))
	if core[eventType] then
		core[eventType](self, eventType, ...)
	end
end

function core:GetSpellInfo(spellID)
	if spellInfo[spellID] then
--~ 		if type(spellInfo[spellID]) == "table" then
--~ 			return spellInfo[spellID].icon, spellInfo[spellID].duration, spellInfo[spellID].debuffType
--~ 		elseif type(spellInfo[spellID]) == "string" then
			local success, icon, duration, debuffType = core:Deserialize(spellInfo[spellID])
			if success then
--~ 				Debug("GetSpellInfo", spellID, icon, duration, debuffType, spellInfo[spellID])
				return icon, duration or 0, debuffType
			end
--~ 		end
	end
	return nil
end

--------------------------------------------------------------------------------------------------
function core:SPELL_AURA_APPLIED(event, _, _, srcGUID, srcName, _, dstGUID, dstName, dstFlags, ...)	--
-- Aura applied, if we know the spell's texture and duration then add it to our buffs list. 	--
--	If not add a question mark the the mob's nameplate.											--
--------------------------------------------------------------------------------------------------
	local spellID, spellName, spellSchool, auraType, amount  = ...
--~ 	Debug(event, spellID, spellName, spellInfo[spellID])
	if spellInfo[spellID] then
		local icon, duration, debuffType = self:GetSpellInfo(spellID)
		if not icon then
			return false
		end
		
		
		local getTime = GetTime()

		guidBuffs[dstGUID] = guidBuffs[dstGUID] or {}
		local count = table_getn(guidBuffs[dstGUID])
		if P.spellOpts[spellName] and P.spellOpts[spellName].show then
			if P.spellOpts[spellName].show == 1 or (P.spellOpts[spellName].show == 2 and srcGUID == playerGUID) then --fix this
				if count == 0 then
					local i = count + 1
					table_insert(guidBuffs[dstGUID], i, {
						name		= spellName,
						icon		= icon,
						duration	= duration or 0,
						playerCast	= srcGUID == playerGUID and 1,
						stackCount	= 1,
						startTime	= getTime,
						expirationTime = getTime + (duration or 0),
						sID = spellID,
						caster = srcName,
					})
					
					if auraType == "DEBUFF" then
						guidBuffs[dstGUID][i].isDebuff = true
						guidBuffs[dstGUID][i].debuffType = debuffType or "none"
					end
				else
					self:RemoveOldSpells(dstGUID)
					count = table_getn(guidBuffs[dstGUID])
					for i=1, count do 
--~ 						if not guidBuffs[dstGUID] or not guidBuffs[dstGUID][i] or not guidBuffs[dstGUID][i].sID then
--~ 							Debug("AURA_APPLIED", guidBuffs[dstGUID], guidBuffs[dstGUID] and guidBuffs[dstGUID][i], guidBuffs[dstGUID] and guidBuffs[dstGUID][i] and guidBuffs[dstGUID][i].sID)
--~ 						end
						if guidBuffs[dstGUID][i].sID == spellID and (not guidBuffs[dstGUID][i].caster or guidBuffs[dstGUID][i].caster == srcName) then
							--I know 2 of the same buff can be on someone, but how do I confirm that?
							-- unitCaster returns a unitID and combatlog has names. =/
							break
						elseif i == count then
							table_insert(guidBuffs[dstGUID], i+1, {
								name		= spellName,
								icon		= icon,
								duration	= (duration or 0),
								playerCast	= srcGUID == playerGUID and 1,
								stackCount	= 1,
								startTime	= getTime,
								expirationTime = getTime + (duration or 0),
								sID = spellID,
								caster = srcName,
							})
							
							if auraType == "DEBUFF" then
								guidBuffs[dstGUID][i+1].isDebuff = true
								guidBuffs[dstGUID][i+1].debuffType = debuffType or "none"
							end
						end
					end
				end
			end
		else
			if auraType == "BUFF" and P.defaultBuffShow == 1 or (P.defaultBuffShow == 2 and srcGUID == playerGUID) 
			or auraType == "DEBUFF" and P.defaultDebuffShow == 1 or (P.defaultDebuffShow == 2 and srcGUID == playerGUID) then

				if count == 0 then
					local i = count + 1
					table_insert(guidBuffs[dstGUID], i, {
						name		= spellName,
						icon		= icon,
						duration	= (duration or 0),
						playerCast	= srcGUID == playerGUID and 1,
						stackCount	= 1,
						startTime	= getTime,
						expirationTime = getTime + (duration or 0),
						sID = spellID,
						caster = srcName,
					})
					
					if auraType == "DEBUFF" then
						guidBuffs[dstGUID][i].isDebuff = true
						guidBuffs[dstGUID][i].debuffType = debuffType or "none"
					end
				else
					for i=1, count do 
						if guidBuffs[dstGUID][i].sID == spellID and (not guidBuffs[dstGUID][i].caster or guidBuffs[dstGUID][i].caster == srcName) then
							--I know 2 of the same buff can be on someone, but how do I confirm that?
							-- unitCaster returns a unitID and combatlog has names. =/
							break
						elseif i == count then
							table_insert(guidBuffs[dstGUID], i+1, {
								name		= spellName,
								icon		= icon,
								duration	= (duration or 0),
								playerCast	= srcGUID == playerGUID and 1,
								stackCount	= 1,
								startTime	= getTime,
								expirationTime = getTime + (duration or 0),
								sID = spellID,
								caster = srcName,
							})
							
							if auraType == "DEBUFF" then
								guidBuffs[dstGUID][i+1].isDebuff = true
								guidBuffs[dstGUID][i+1].debuffType = debuffType or "none"
							end
						end
					end
				end
			end
		end

--~ 		Debug("Combatlog", dstName, spellName, self:UpdatePlateByGUID(dstGUID), self:FlagIsPlayer(dstFlags))
		if not self:UpdatePlateByGUID(dstGUID) and self:FlagIsPlayer(dstFlags) then
			local shortName = core:RemoveServerName(dstName)
			nametoGUIDs[shortName] = dstGUID
			self:UpdatePlateByName(shortName)
		end
	--[[else --unknown spellID, add a question mark.
		if guidBuffs[dstGUID] then
			local count = table_getn(guidBuffs[dstGUID])
			local getTime = GetTime()
			
			if auraType == "BUFF" and P.defaultBuffShow == 1 or (P.defaultBuffShow == 2 and srcGUID == playerGUID) 
			or auraType == "DEBUFF" and P.defaultDebuffShow == 1 or (P.defaultDebuffShow == 2 and srcGUID == playerGUID) then
			
				if count == 0 then
					local i = count + 1
					table_insert(guidBuffs[dstGUID], i, {
						name		= spellName,
						icon		= core.unknownIcon,
						duration	= 0,
						playerCast	= srcGUID == playerGUID and 1,
						stackCount	= 1,
						startTime	= getTime,
						expirationTime = 0,
						sID = spellID,
						caster = srcName,
					})
					
					if auraType == "DEBUFF" then
						guidBuffs[dstGUID][i].isDebuff = true
						guidBuffs[dstGUID][i].debuffType = "none"
					end
				else
					for i=1, count do 
						if guidBuffs[dstGUID][i].sID == spellID and (not guidBuffs[dstGUID][i].caster or guidBuffs[dstGUID][i].caster == srcName) then
							--I know 2 of the same buff can be on someone, but how do I confirm that?
							-- unitCaster returns a unitID and combatlog has names. =/
							break
						elseif i == count then
							table_insert(guidBuffs[dstGUID], i+1, {
								name		= spellName,
								icon		= core.unknownIcon,
								duration	= 0,
								playerCast	= srcGUID == playerGUID and 1,
								stackCount	= 1,
								startTime	= getTime,
								expirationTime = 0,
								sID = spellID,
								caster = srcName,
							})
							
							if auraType == "DEBUFF" then
								guidBuffs[dstGUID][i+1].isDebuff = true
								guidBuffs[dstGUID][i+1].debuffType = "none"
							end
						end
					end
				end
			end
		end]]
	end
	
end

------------------------------------------------------------------------------------------
function core:SPELL_AURA_REMOVED(event, _, _, srcGUID, srcName, _, dstGUID, dstName, dstFlags, ...)	--
-- Spell's been removed, remove it from our table.										--
------------------------------------------------------------------------------------------
	local spellID, spellName, spellSchool, auraType  = ...
	
	if guidBuffs[dstGUID] then
		for i=table_getn(guidBuffs[dstGUID]), 1, -1 do 
			if guidBuffs[dstGUID][i].sID == spellID and (not guidBuffs[dstGUID][i].caster or guidBuffs[dstGUID][i].caster == srcName) then
--~ 					Debug("REMOVED 3", spellName, srcName, dstName)
				table_remove(guidBuffs[dstGUID], i)
				if not self:UpdatePlateByGUID(dstGUID) and self:FlagIsPlayer(dstFlags) then
					self:UpdatePlateByName(dstName)
				end
				return
			end
		end
	end
end


function core:SPELL_AURA_REFRESH(event, _, _, srcGUID, srcName, _, dstGUID, dstName, dstFlags, ...)
	local spellID, spellName, spellSchool, auraType  = ...
	if guidBuffs[dstGUID] then
		for i=table_getn(guidBuffs[dstGUID]), 1, -1 do 
			if guidBuffs[dstGUID][i].sID == spellID and (not guidBuffs[dstGUID][i].caster or guidBuffs[dstGUID][i].caster == srcName) then
				local getTime = GetTime()
				guidBuffs[dstGUID][i].startTime	= getTime
				guidBuffs[dstGUID][i].expirationTime = getTime + guidBuffs[dstGUID][i].duration
--~ 				Debug("AURA_REFRESH", srcName, spellName, dstName)
				
				if not self:UpdatePlateByGUID(dstGUID) and self:FlagIsPlayer(dstFlags) then
					self:UpdatePlateByName(dstName)
				end
				return
			end
		end
	end
	--Spell isn't in our list, let's add it.
	self:SPELL_AURA_APPLIED(event, _, _, srcGUID, srcName, _, dstGUID, dstName, dstFlags, ...)
end

--DOSE = spell stacking
function core:SPELL_AURA_APPLIED_DOSE(event, _, _, srcGUID, srcName, _, dstGUID, dstName, dstFlags, ...)
	local spellID, spellName, spellSchool, auraType  = ...

--~ 	Debug("DOSE", srcName, spellName, dstName)
	if spellInfo[spellID] and guidBuffs[dstGUID] then
		for i=table_getn(guidBuffs[dstGUID]), 1, -1 do 
			if guidBuffs[dstGUID][i].sID == spellID and (not guidBuffs[dstGUID][i].caster or guidBuffs[dstGUID][i].caster == srcName) then
--~ 				Debug("SPELL_AURA_APPLIED_DOSE", srcName, spellName, dstName)
				guidBuffs[dstGUID][i].stackCount = guidBuffs[dstGUID][i].stackCount + 1
				
				if not self:UpdatePlateByGUID(dstGUID) and self:FlagIsPlayer(dstFlags) then
					self:UpdatePlateByName(dstName)
				end
				return
			end
		end
	end
	--Spell isn't in our list, let's add it.
	--Note this event could have fired on the 5th stack but our spell frame will only show it applied once. 
	self:SPELL_AURA_APPLIED(event, _, _, srcGUID, srcName, _, dstGUID, dstName, dstFlags, ...)
end

--sheep break
-- Gouge break
function core:SPELL_AURA_BROKEN_SPELL(event, _, _, srcGUID, srcName, _, dstGUID, dstName, dstFlags, ...)
--~ 	local spellID, spellName, spellSchool, auraType  = ...
	self:SPELL_AURA_REMOVED(event, _, _, srcGUID, srcName, _, dstGUID, dstName, dstFlags, ...)
end

--Not sure when this fires.
-- fires on sap break
function core:SPELL_AURA_BROKEN(event, _, _, srcGUID, srcName, _, dstGUID, dstName, dstFlags, ...)
--~ 	local spellID, spellName, spellSchool, auraType  = ...
	self:SPELL_AURA_REMOVED(event, _, _, srcGUID, srcName, _, dstGUID, dstName, dstFlags, ...)
end


function core:UNIT_DIED(event, _, _, srcGUID, srcName, _, dstGUID, dstName, dstFlags, ...)
	if guidBuffs[dstGUID] then
		--Remove all known buffs for that person. Maybe we're in a BG and don't need their old buffs on our plates.
		for i=table_getn(guidBuffs[dstGUID]), 1, -1 do 
			table_remove(guidBuffs[dstGUID], i)
		end
		if not self:UpdatePlateByGUID(dstGUID) and self:FlagIsPlayer(dstFlags) then
			self:UpdatePlateByName(dstName)
		end
	end
end

function core:UNIT_DESTROYED(event, _, _, srcGUID, srcName, _, dstGUID, dstName, dstFlags, ...)
	self:UNIT_DIED(event, _, _, srcGUID, srcName, _, dstGUID, dstName, dstFlags, ...)
end

function core:UNIT_DISSIPATES(event, _, _, srcGUID, srcName, _, dstGUID, dstName, dstFlags, ...)
	self:UNIT_DIED(event, _, _, srcGUID, srcName, _, dstGUID, dstName, dstFlags, ...)
end

function core:PARTY_KILL(event, _, _, srcGUID, srcName, _, dstGUID, dstName, dstFlags, ...)
	self:UNIT_DIED(event, _, _, srcGUID, srcName, _, dstGUID, dstName, dstFlags, ...)
end


function core:FlagIsPlayer(flags)
	if bit_band(flags, COMBATLOG_OBJECT_TYPE_PLAYER) == COMBATLOG_OBJECT_TYPE_PLAYER then
		return true
	end
	return nil
end
